package de.mutschler.rpdl.calc;

import de.mutschler.util.*;
import de.mutschler.rpdl.*;
import java.util.*;

/**
 * This class represents the event handler in RPDL.
 * You can queue the events (duplicate events are removed)
 * and then call the execute() to execute all events.
 * Note: The Handler checks for cycles and other stuff.
 */

public class CalcHandler extends LogObj
{
	/** store the RpdlBase object. */
	private RpdlBase rpdlMain;
	private RpdlBase getMain() {return rpdlMain;}

	//----------------------------------------------------------------------
	//------- Constructor(s)
	//----------------------------------------------------------------------

	public CalcHandler(RpdlBase main) {
		rpdlMain = main;
	}
	
	//----------------------------------------------------------------------
	//------- Calculation
	//----------------------------------------------------------------------

	/** A list of events. This makes sure, that there are only
	 * events of a different type. e.g. the same attribute
	 * should not be calculated twice!
	 */
	private List queuedCalculations = new LinkedList();
	
	public int ignoreCnt = 0;
	public int calcCnt = 0;
	public int calcTime = 0;
	
	/** Add an event to the queue. */
	public void addCalcItem(CalcItem ci) throws RpdlException {
		// do not add this object, if it already exists in the queue
		synchronized(queuedCalculations) {
/*			CalcItem found = null;
			for (Iterator it=queuedCalculations.iterator(); it.hasNext(); ) {
				found = (CalcItem)it.next();
				if (ci.equals(found)) break;
				found = null;
			}
			if (found != null) {
			}
*/
			if (!queuedCalculations.contains(ci)) {
				calcCnt++;
				debug("adding calc '" + ci + "'");
				if (performedCalculations.contains(ci)) {
					throw new RpdlException(
						"calculation " + ci + " has already been performed. possible cycle." + CR +
						"the other calculation is " + ((CalcItem)performedCalculations.get(performedCalculations.indexOf(ci))));
				}
				queuedCalculations.add(ci);
			}
			else {
				debug("didn't add '" + ci + "' it already exists.");
				ignoreCnt++;
			}
		}
	}

	/**
	 * Remove a calcItem from the list of queued calculations.
	 * This is useful, e.g If you remove an attribute and you have a queued
	 * calculation for this attribute.
	 */
	public void removeCalcItem(CalcItem ci) {
		boolean res = queuedCalculations.remove(ci);
		debug("removing calculation " + ci + ": " + (res?"removed":"none found"));
	}

	/**
	 * Suspend the execution. If you reenable this, then all the queued
	 * calculations are performed. So, suspend(false) will always execute
	 * @param suspend: if true, execution is suspended, if false exeution is reenabled.
	 */
	public void suspend(boolean suspend) throws RpdlException {
		this.suspend = suspend;
		if (!suspend)
			execute();
	}
	
	private boolean suspend = false;

	/**
	 * While executing, all the finished calculations will
	 * be moved into this list. Therefore we can check on  addCalcObject()
	 * if an event is created twice, and check this way for cycles
	 * during the calculation. Note: this should not happen, but who knows...
	 */
	private List performedCalculations = new LinkedList();
	
	/** execute the queued events. */
	public synchronized void execute() throws RpdlException {
		if (suspend) return;	// no execution if suspended
		debug("performing # " + queuedCalculations.size() +" calculations ....");
		if (queuedCalculations.isEmpty()) return;
		long starttime = System.currentTimeMillis();

		// first we must make sure we have the "first" calculation
		// therefore we have to take care of the dependencies of the
		// calculations. 

		// loop while we have calculations to perform
		while(!queuedCalculations.isEmpty()) {
			// get the first element of the queue, and put it in the done queue
			// then execute the calculation
			CalcItem ci;
			synchronized(queuedCalculations) {	// just to synchronize with addCalcObj()
				ci = (CalcItem)queuedCalculations.get(0);
				// check if we have an element that should be executed before this one
				for (int i=1; i<queuedCalculations.size(); i++) {
					CalcItem ci2 = (CalcItem)queuedCalculations.get(i);
					if (ci.compareTo(ci2) > 0) {
						debug("using '" + ci2 + "' before of '" + ci + "'");
						ci = ci2;
					}
				}
				queuedCalculations.remove(ci);
				performedCalculations.add(ci);
			}
			// now execute the calculation
			try {
				ci.execute(getMain());
			} catch(Exception e) {
				error("Failed execution on calculation '" + ci + "':", e);
				getMain().addError(e);
			}
		} // end while list not empty
		
		synchronized(queuedCalculations) {
			performedCalculations.clear();
		}
		long endtime = System.currentTimeMillis();
		calcTime += (endtime-starttime);
	}
}