package de.mutschler.rpdl.entity;

import de.mutschler.util.*;
import de.mutschler.rpdl.*;
import de.mutschler.rpdl.statement.*;
import java.util.*;

/**
 * This Class is the superclass of all Objects and type.
 * It contains basic support for a list of commands.
 */

public abstract class Entity extends NamedObj
{
	//----------------------------------------------------------------------
	//------- Attributes
	//----------------------------------------------------------------------

	/** Pointer to the rpdl base. */
	private RpdlBase main;
	
	public RpdlBase getMain() {return main;}
	//----------------------------------------------------------------------
	//------- Constructor(s)
	//----------------------------------------------------------------------
	
	/** Create a new entity, and specify its name. */
	public Entity(RpdlBase rbase, String s, Set types) {
		super(s);
		main = rbase;
		parentTypes = types;
		if (types == null) parentTypes = new HashSet();
	}

	/**
	 * Create a plain copy from another object. The parent types and statements
	 * are copied (only thier reference).
	 */
	public Entity(Entity orig, RpdlBase base, String s) throws RpdlException {
		this(base, s, null);
		// now copy the parent types:
		for (Iterator it = orig.parentTypes(); it.hasNext();) {
			addType((RpdlType)it.next());
		}
		// copy the statements
		for (Iterator it = orig.getStatements(); it.hasNext();) {
			// cannot use addStatement, because setDerivedOrigin() would cause an exception
			stmtList.add((Stmt)it.next());
		}
		
	}

	//----------------------------------------------------------------------
	//------- own defined Statements
	//----------------------------------------------------------------------
	
	/** List of statements owned by this entity */
	private List stmtList = new LinkedList();

	/** Add a statement to the internal lists.
	 */
	public void addStatement(Stmt s) throws RpdlException {
		ddebug("Adding statement: " + s);
		// so the statement has got a link to its entity
		s.setDerivedEntity(this);
		stmtList.add(s);
	}

	/** Get all statement-Items of this object */
	public Iterator getStatements() {return stmtList.iterator();}

	//----------------------------------------------------------------------
	//------- parent Type handling
	//----------------------------------------------------------------------
	
	/** List of Types, which this entity is derived from. */
	private Set parentTypes;
	
	/** This method checks, if this entity is of the same type. Note:
	 * the type is recursively scanned through the typeMap!
	 */
	public boolean isType(RpdlType type) {return isType(type.getName());}

	public boolean isType(String name) {
		for (Iterator it = parentTypes.iterator(); it.hasNext(); ) {
			RpdlType rt = (RpdlType)it.next();
			// check if the type has the same name
			if (name.equals(rt.getName())) return true;
			// otherwise do the recursion
			if (rt.isType(name)) return true;
		}
		return false;
	}
	
	
	
	/** Get the parent types */
	public Iterator parentTypes() {return parentTypes.iterator();}

	/** So we can do manipulations on the parents in a subclass. */
	protected Set getParentTypesSet() {return parentTypes;}

	/** Add another type to the list. If the type already exists,
	 * nothing is done.
	 * @return true if the new type has not been there before.
	 */
	public boolean addType(RpdlType type) throws RpdlException {
		boolean res = !isType(type.getName());
		parentTypes.add(type);
		return res;
	}
	
	/** Add another type to the list. If the type already exists,
	 * nothing is done.
	 * @return true, if the entity is not of this type anymore.
	 */
	public boolean removeType(RpdlType type) {
		parentTypes.remove(type);
		return !isType(type.getName());
	}
	
	/**
	 * put the parents in the form
	 * + sub1
	 * | + sub11
	 * | + sub12
	 * + sub2
	 *   + sub21
	 */
	public String parentTypesString() {
		TreeFormatter tf = new TreeFormatter() {
			public String getInfo(Object o)		{return ((Entity)o).getName();}
			public Iterator subitems(Object o)	{return ((Entity)o).parentTypes();}
		};
		return tf.getSubTreeString(this);
	}

	//----------------------------------------------------------------------
	//------- Overridden system methods
	//----------------------------------------------------------------------
	
	/** The Hash for an object is the hash of its name. */
	public int hashCode() {return getName().hashCode();}

	public String toString() {
		StringBuffer sb = new StringBuffer(200);
		sb.append(getName());
		// the parent types
		boolean start = true;
		for (Iterator it=parentTypes.iterator(); it.hasNext(); ) {
			if (start)	sb.append(" FROM ");
			else		sb.append(", ");
			start = false;
			sb.append(((RpdlType)it.next()).getName());
		}
		// now the statements
		sb.append(CR + "{" + CR);
		for (Iterator it = getStatements(); it.hasNext(); ) {
			sb.append(StringUtil.indent(((Stmt)it.next()).toString()));
			sb.append(CR);
		}
		sb.append("}");
		sb.append(CR + CR + "Parenttypes:" + CR + parentTypesString() + CR);
		return sb.toString();
	}
}