package de.mutschler.rpdl.entity;

import java.util.*;
import de.mutschler.util.*;
import de.mutschler.rpdl.basicvalue.BasicValue;
import de.mutschler.rpdl.basicvalue.ValueObject;
import de.mutschler.rpdl.*;

/**
 * This class represents an access to an attribute. It simply reads
 * the value of the attribute represented by the object
 * specified in objectaccess.
 */

public class FormulaAttribute extends RpdlFormula implements java.io.Serializable
{
	static final long serialVersionUID = -1893849280184108557L;
	//----------------------------------------------------------------------
	//------- Attributes
	//----------------------------------------------------------------------
	/** The name of the attribute. */
	private String attrName;
	/**
	 * The path to the actual object, which this attribute belongs to.
	 * This formula must return type OBJECT. Note: this can be null!
	 */
	private RpdlFormula object;
	
	//----------------------------------------------------------------------
	//------- Constructor(s)
	//----------------------------------------------------------------------
	
	/**
	 * This constructur requires an object to refer to, and
	 * a name of the specific attribute.
	 */
	public FormulaAttribute(RpdlFormula oa, String name) {
		object = oa;
		attrName = name;
	}
	
	/**
	 * This constructur requires an object to refer to, and
	 * a name of the specific attribute.
	 */
	public FormulaAttribute(String name) {
		// an empty objectaccess means the current object
		this(null, name);
	}

	//----------------------------------------------------------------------
	//------- Accessors
	//----------------------------------------------------------------------
	
	/** Get the name of this attribute. */
	public String getName() {return attrName;}
	
	//----------------------------------------------------------------------
	//------- Methods for interface RpdlFormula
	//----------------------------------------------------------------------
	
	/**
	 * Perform a calculation, and return the result.
	 */
	public BasicValue calc(RpdlBase main, RpdlObject obj) throws RpdlException {
		Set s = getDependencies(main, obj);
		if (s.size() != 1) {
			throw new RpdlException("Only one attribute can be calculated. got #" + s.size());
		}
		RpdlAttribute attr = (RpdlAttribute)s.iterator().next();
		return attr.getValue();
	}

	/**
	 * Get the type of result.
	 */
	public String getResultType(RpdlBase main, Entity e) throws RpdlException {
		RpdlObject obj = null;
		if (object != null) {
			if (!ValueObject.TYPENAME.equals(object.getResultType(main, e)))
				throw new RpdlException("Subfunction does not return an object", getOrigin());
		}
		try {
			if (e instanceof RpdlObject) {
				obj = getObject(main, (RpdlObject)e);
			}
			RpdlAttribute attr = obj.getAttribute(getName());
			if (attr == null) {
				throw new RpdlException("Attribute " + getName() + " not found in object '" + obj + "'");
			}
			return attr.getType();
		}
		catch (Exception ex) {	// by default get the type from the attribute definition
			RpdlAttributeDef attrDef =  main.getAttributeDef(getName());
			if (attrDef == null) {
				throw new RpdlException("Attribute " + getName() + " not defined!");
			}
			return attrDef.getType();
		}
	}
	
	/**
	 * Find the right object where to get the attribute from.
	 * This method takes care of object == null, in which case it returns obj
	 */
	private RpdlObject getObject(RpdlBase main, RpdlObject obj) throws RpdlException {
		if (object == null) return obj;
		ValueObject vo = (ValueObject)object.calc(main, obj);
		return vo.getValue();
	}

	/**
	 * Get a list of all dependencies for this formula.
	 * This returns a Set with objects of type RpdlAttribute
	 */
	public Set getDependencies(RpdlBase main, RpdlObject obj) throws RpdlException {
		Set s = new HashSet();
		if (object instanceof FormulaObject) {
			Set so = ((FormulaObject)object).getObjects(main, obj);
			ddebug("#" + so.size() + " dependencies from FormulaObject.");
			for (Iterator it=so.iterator(); it.hasNext();) {
				RpdlObject ro = (RpdlObject)it.next();
				RpdlAttribute a = ro.getAttribute(getName());
				if (a != null) {
					s.add(a);
				}
				else {
					throw new RpdlException("Attribute " + getName() + " not found in object "+ ro, getOrigin());
				}
			}
		}
		else {
		// by default use the returned object
			RpdlObject ro = getObject(main, obj);
			RpdlAttribute a = ro.getAttribute(getName());
			if (a != null) {
				s.add(a);
			}
			else {
				throw new RpdlException("Attribute " + getName() + " not found in object "+ ro, getOrigin());
			}
		}
		return s;
	}

	//----------------------------------------------------------------------
	//------- Overridden system methods
	//----------------------------------------------------------------------
	
	public String toString() {
		String obj = (object==null?"":object.toString());
		StringBuffer sb = new StringBuffer(20);
		sb.append("(Attr: ");
		if (!StringUtil.isEmpty(obj)) {
			sb.append(obj).append('.');
		}
		sb.append(attrName.toString()).append(')');
		return  sb.toString();
	}
	
}
