package de.mutschler.rpdl.fileimport;

import de.mutschler.util.LogObj;
import de.mutschler.rpdl.RpdlBase;
import de.mutschler.rpdl.command.RpdlCommand;
import java_cup.runtime.*;
import java.io.*;
import de.mutschler.util.Log;

/**
 * This class invokes the parser, and the parser sends all commands through
 * the addCommand() method.
 *
 * There is a caching implemented, which checks for a file with the extension
 * ".cache", and if it is newer that the original file, then the cache file
 * is used for reading the commands. It consists of the serialized objects
 * from the addCommand() method.
 */

public class FileImporter extends LogObj
{
	//----------------------------------------------------------------------
	//------- Attributes
	//----------------------------------------------------------------------

	/** The file to read. */
	private File file;

	/** The RpdlBase to fill... */
	private RpdlBase rBase;
	
	//----------------------------------------------------------------------
	//------- Constructor(s)
	//----------------------------------------------------------------------

	/** Read the current file. */
	public FileImporter(RpdlBase rb, File f) throws RpdlScanException {
		init(rb, f);
	}
	
	/** Read a file relative to the parent FileImporter.
	 * This method tries to read the file directly, or relative
	 * to the path of the previous file.
	 */
	public FileImporter(RpdlBase rb, FileImporter fi, File f) throws RpdlScanException {
		if (!f.isFile()) {
			f = new File(fi.getFile().getParent(), f.getPath());
		}
		init(rb, f);
	}

	private void init(RpdlBase rb, File f) throws RpdlScanException {
		file = f;
		if (!f.isFile()) {
			throw new RpdlScanException("File \"" + f.getPath() + "\" not found!");
		}
		rBase = rb;
	}
	
	//----------------------------------------------------------------------
	//------- Accessors
	//----------------------------------------------------------------------
	
	/** Get the filename. */
	public File getFile() {return file;}
	
	/** Get the RpdlBase. */
	public RpdlBase getMain() {return rBase;}

	//----------------------------------------------------------------------
	//------- Methods
	//----------------------------------------------------------------------
	
	/**
	 * This is the main method to actually start the file import.
	 * It first tries to read the data from the cache, otherwise the cache
	 * is created.
	 */
	public void scanfile () {
		// uncomment the following line, to enable caching
		//if (readFromCache()) return;
		debug("Scanning file " + getFile() + "...");
		Symbol tree = null;
		try {
			RpdlParser parser = new RpdlParser(this, new RpdlScanner(new FileReader(file)));
			parser.parse();
			closeCacheFile();
		} catch (Exception e) {
			error("Parse-error"+ e);
			Log.exception(e);
			getMain().addError(e);
			cacheErrorCleanup();	// in the case of an error, delete the cache
		}
	}
	
	/** include the new file */
	public void include(String file) {
		try {
			if (cacheOut != null) {
				try {
					cacheOut.writeObject(new CmdInclude(file));
				} catch (IOException e) {
					error("Couldn't write into the cache file " + cacheFile, e);
					cacheErrorCleanup();
				}
			}
			FileImporter fi = new FileImporter(getMain(), this, new File(file));
			fi.scanfile();
		} catch (Exception e) {
			error("Parse-error"+ e);
			Log.exception(e);
			getMain().addError(e);
		}
	}
	
	//----------------------------------------------------------------------
	//------- Caching
	//----------------------------------------------------------------------

	/** The extension of the cachefile. appended to the original file. */
	private static final String CACHE_EXTENSION=".cache";
	
	/** version of the cachefile. This way we can easily change the format.
	 * of the cache file. It is written as the second item.
	 */
	private static final int CACHE_VERSION=1;
	
	/** This string must be the first in a cache file. */
	private static final String CACHE_IDENTIFIER="RPDL cachefile";
	
	/** If this is not null, then write all the commands to the stream. */
	private ObjectOutputStream cacheOut = null;
	private File cacheFile = null;
	
	/** All commands must go through this method. it writes the cache. */
	public void addCommand(RpdlCommand cmd) {
		if (cacheOut != null) {
			try {
				cacheOut.writeObject(cmd);
			} catch (IOException e) {
				error("Couldn't write into the cache file " + cacheFile, e);
				cacheErrorCleanup();
			}
		}
		rBase.addCommand(cmd);
	}

	/**
	 * Do the cleanup. This writes the identifier at the end of the file,
	 * and closes it.
	 */
	private void closeCacheFile() {
		if (cacheOut == null) return;	// no cache file currently writtem
		try {
			cacheOut.writeObject(CACHE_IDENTIFIER);
			cacheOut.close();
			debug("successfully written cachefile '" + cacheFile + "'");
		} catch (Exception e) {
			error("Failed closing cachefile '" + cacheFile + "'.", e);
			cacheErrorCleanup();
		}
	}

	private void cacheErrorCleanup() {
		if (cacheOut == null) return;
		try {
			// in the case of an error, close the stream and delete the cache file
			cacheOut = null;
			cacheOut.close();
			cacheFile.delete();
		} catch (Exception e) {
			error("While cleanup: ", e);
		}
	}
	/**
	 * This method tries to read from the cache file. If the data could be
	 * read from the cache file, true is returned.
	 */
	private boolean readFromCache() {
		cacheFile = new File(getFile().getPath() + CACHE_EXTENSION);
		try {
			if (!cacheFile.exists()) {
				debug("Cachefile doesn't exist for file '" + getFile() + "'");
				setupCacheFile();
				return false;
			}
			// we have a cachefile. now check if it is newer
			if (getFile().lastModified() > cacheFile.lastModified()) {
				debug("The cache of file '" + getFile() + "' is outdated.");
				setupCacheFile();
				return false;
			}
			debug("reading data from cachefile '" + cacheFile + "'");
			// we have a uptodate cachefile. so read the ident and version
			ObjectInputStream oi = new ObjectInputStream(new FileInputStream(cacheFile));
			try {
				byte [] id = new byte[CACHE_IDENTIFIER.length()];
				oi.read(id);
				String ident = new String(id);
				// check the identifier
				if (!CACHE_IDENTIFIER.equals(ident)) {
					error("file '" + cacheFile + "' is not a valid RPDL cachefile. overwriting it.");
					setupCacheFile();
					return false;
				}
				// check the version
				int cacheVersion = oi.readInt();
				if (cacheVersion != CACHE_VERSION) {
					debug("file '" + cacheFile + "' is outdated.");
					setupCacheFile();
					return false;
				}
				
				// finally we can read the objects
				Object obj;
				while((obj = oi.readObject()) != null) {
					// if we have an object of the identifier, we have finished reading
					if (obj instanceof String) {
						String end = (String)obj;
						if (CACHE_IDENTIFIER.equals(end)) {
							debug("successfully read data from cachefile");
							// close everything
							oi.close();
							return true;
						}
					}
					// check for include
					if (obj instanceof CmdInclude) {
						CmdInclude ci = (CmdInclude)obj;
						include(ci.file);
						continue;
					}
					// by default add the command to the RpdlBase
					addCommand((RpdlCommand)obj);
				}
				error("got null-object from input");
				oi.close();
				return false;
			} catch (Exception e) {
				error("While reading from the cache file '" + cacheFile +"'. Trying to write new cache file.", e);
				oi.close();
				setupCacheFile();
				return false;
			}
		} catch (Exception e) {
			error("Failed setting up the cache file " + cacheFile, e);
			cacheErrorCleanup();
			return false;
		}
	}
	
	/** Initialize the caching. Setup the objectOutputStream and
	 *  write the header (CACHE_IDENTIFIER & CACHE_VERSION).
	 */
	private void setupCacheFile() throws IOException {
		debug("setting up cachefile '" + cacheFile + "'");
		cacheOut = new ObjectOutputStream(new FileOutputStream(cacheFile));
		// write the header
		cacheOut.writeBytes(CACHE_IDENTIFIER);
		cacheOut.writeInt(CACHE_VERSION);
	}
	
	/** Display an error. Used from the parser. */
	public void import_error(String message, int line, int col)
	{
		String msg = "Error in file " + getFile().getPath();
		if (line != -1) msg = msg + " line " + line;
//		if (col != -1)  msg = msg + " col " + col;
		msg = msg + ": " + message;
		error(msg);
		getMain().addError(msg);
	}
}

/**
 * This is simply a class for the cache file. If this
 * kind of object is in the cache, the include the given file.
 */
class CmdInclude implements Serializable {
	static final long serialVersionUID = 8348189097385745039L;
	String file;
	public CmdInclude(String file) {
		this.file=file;
	}
		
}
	
