/* This is a Scanner for the RPDL language. */

package de.mutschler.rpdl.fileimport;
import java_cup.runtime.*;
import de.mutschler.util.Log;

%%


%line
//%column
%cup
%8bit
%ignorecase
%public
%class RpdlScanner
%implements sym
%state COMMENT, HISTORY, HISTSTR, QSTRING, EOLCOMMENT

%{
	private Symbol symbol(int type) {
		return symbol(type, null);
	}
	private Symbol symbol(int type, Object val) {
//		ddebug("got symbol: "+ type + " value= " + yytext());
		return new Symbol(type, yyline+1, 0, val);
	}

	public int getLine() {return yyline+1;}
	/** a bit logging support */	
	private final String _clsName = getClass().getName();
	private void debug(String s)  {Log.log(Log.DEBUG,  _clsName, s);}
	private void ddebug(String s) {Log.log(Log.DDEBUG, _clsName, s);}
	
%} 

_ALPHA=[A-Za-z_]
DIGIT=[0-9]
Identifier={_ALPHA}({_ALPHA}|{DIGIT})*
HISTBLOCK=([^\n\[]|[^\n]\[|\n[^\[])*
DICEVAL=(4|6|8|10|12|20|30|100)
DICESV={DIGIT}+[WwDd]{DICEVAL}
DICES={DICESV}({WHITE_SPACE_CHAR}*[+-]{WHITE_SPACE_CHAR}*({DICESV}|{DIGIT}))*

NEWLINE=\r|\n|\r\n
WHITE_SPACE_CHAR=[\n\r\ \t\b\012]

%% 

<YYINITIAL> {
  "+" { return (symbol(PLUS)); }
  "-" { return (symbol(MINUS)); }
  "*" { return (symbol(TIMES)); }
  "/" { return (symbol(DIVIDE)); }
  "(" { return (symbol(LPAREN)); }
  ")" { return (symbol(RPAREN)); }
  ";" { return (symbol(SEMICOLON)); }
  ":" { return (symbol(COLON)); }
  "," { return (symbol(COMMA)); }
  "." { return (symbol(DOT)); }
  "#" { return (symbol(CROSS)); }
  "$" { return (symbol(DOLLAR)); }
  ".." { return (symbol(DOTDOT)); }
  "[" { return (symbol(LBRACKET)); }
  "]" { return (symbol(RBRACKET)); }
  "{" { return (symbol(LBRACE)); }
  "}" { return (symbol(RBRACE)); }
  ">" { return (symbol(GT)); }
  ">=" { return (symbol(GE)); }
  "<" { return (symbol(LT)); }
  "<=" { return (symbol(LE)); }
  "=" { return (symbol(EQ)); }
  "!=" { return (symbol(NE)); }
  "<>" { return (symbol(NE)); }
  "+=" { return (symbol(ADDTO)); }

  "\"" {yybegin(QSTRING);}

  "COMMENT:START" {yybegin(HISTORY); return symbol(HIST_START); }

  "IF"		{return symbol(KW_IF, yytext()); }
  "ON"		{return symbol(KW_ON, yytext()); }
  "TO"		{return symbol(KW_TO, yytext()); }
  "ADD"		{return symbol(KW_ADD, yytext()); }
  "INT"		{return symbol(KW_INT, yytext()); }
  "SET"		{return symbol(KW_SET, yytext()); }
  "BASE"	{return symbol(KW_BASE, yytext()); }
  "BOOL"	{return symbol(KW_BOOL, yytext()); }
  "DATE"	{return symbol(KW_DATE, yytext()); }
  "DICE"	{return symbol(KW_DICE, yytext()); }
  "ELSE"	{return symbol(KW_ELSE, yytext()); }
  "FROM"	{return symbol(KW_FROM, yytext()); }
  "LINK"	{return symbol(KW_LINK, yytext()); }
  "TYPE"	{return symbol(KW_TYPE, yytext()); }
  "WARN"	{return symbol(KW_WARN, yytext()); }
  "ERROR"	{return symbol(KW_ERROR, yytext()); }
  "TABLE"	{return symbol(KW_TABLE, yytext()); }
  "TITLE"	{return symbol(KW_TITLE, yytext()); }
  "ASSERT"	{return symbol(KW_ASSERT, yytext()); }
  "CHANGE"	{return symbol(KW_CHANGE, yytext()); }
  "CREATE"	{return symbol(KW_CREATE, yytext()); }
  "MODIFY"	{return symbol(KW_MODIFY, yytext()); }
  "NUMBER"	{return symbol(KW_NUMBER, yytext()); }
  "OBJECT"	{return symbol(KW_OBJECT, yytext()); }
  "STRING"	{return symbol(KW_STRING, yytext()); }
  "UNLINK"	{return symbol(KW_UNLINK, yytext()); }
  "COMMENT" {return symbol(KW_COMMENT, yytext()); }
  "DEFAULT" {return symbol(KW_DEFAULT, yytext()); }
  "INCLUDE"	{return symbol(KW_INCLUDE, yytext()); }
  "LIBRARY"	{return symbol(KW_LIBRARY, yytext()); }
  "SETLAST"	{return symbol(KW_SETLAST, yytext()); }
  "REQUIRE"	{return symbol(KW_REQUIRE, yytext()); }
  "REQUIRED"	{return symbol(KW_REQUIRED, yytext()); }
  "SETFIRST"	{return symbol(KW_SETFIRST, yytext()); }
  "ATTRIBUTE"	{return symbol(KW_ATTRIBUTE, yytext()); }
  "BASICTYPE"	{return symbol(KW_BASICTYPE, yytext()); }

  "/*" { yybegin(COMMENT); }
  "//" { yybegin(EOLCOMMENT); }

/**** constants ****/
  {DIGIT}+				{ return symbol(CONST_INT, new Integer(yytext())); }
  {DIGIT}*\.{DIGIT}+	{ return symbol(CONST_NUMBER, new Float(yytext())); }
  /** Dices must come *after* int. So a plain number creates a CONST_INT */
  {DICES}				{ return symbol(CONST_DICE, yytext()); }
  "true"				{ return symbol(CONST_BOOL, Boolean.TRUE); }
  "false"				{ return symbol(CONST_BOOL, Boolean.FALSE); }

  {Identifier} {return symbol(IDENT, yytext()); }

  {WHITE_SPACE_CHAR}* {}
  . {System.err.println("error matching text");}

}

/* start a quoted string */
<QSTRING> {
	"\"" {yybegin (YYINITIAL);}
	([^\"]| \\\")* {return symbol(QSTR, yytext()); }
}

/* HISTORY state begins with keyword "COMMENT:START" */
<HISTORY> {

  "]" { yybegin(HISTSTR); return (symbol(RBRACKET)); } /* after a rbracket a history string begins */

  "COMMENT:END" { yybegin(YYINITIAL); return symbol(HIST_END);}

  {WHITE_SPACE_CHAR}* {}
  . {System.err.println("error matching text");}
}

/* History-string takes all up to a bracket starting at the beginning of the line */
<HISTSTR> {
  {HISTBLOCK} { return symbol(HIST_STRING, yytext().trim());}
  "\n[" { yybegin(HISTORY); return symbol(LBRACKET); }	/* LBRACKET at the beginning of line */
}

<COMMENT> {
  "*/" { yybegin(YYINITIAL); }
  {WHITE_SPACE_CHAR}* {}
  . {}
//  {COMMENT_TEXT} { }
}

<EOLCOMMENT> {
  {NEWLINE} {yybegin(YYINITIAL);}
  . {}
}

