package de.mutschler.rpdl.statement;

import de.mutschler.rpdl.*;
import de.mutschler.rpdl.entity.*;
import de.mutschler.rpdl.basicvalue.*;

import java.util.*;

public class Stmt_Set extends Stmt_Formula implements java.io.Serializable
{
	static final long serialVersionUID = -4234617827601921700L;
	
	/** High Priority of the statement. */
	public static final int SETPRI_FIRST=-1;
	/** Normal Priority of the statement. */
	public static final int SETPRI_NORM=0;
	/** Low Priority of the statement. */
	public static final int SETPRI_LAST=1;
	
	//----------------------------------------------------------------------
	//------- Attributes
	//----------------------------------------------------------------------
	
	/** The target attribute to write the result to. */
	private FormulaAttribute target;

	/** The priority. */
	private int priority = PRI_NORM;
	
	//----------------------------------------------------------------------
	//------- Constructor(s)
	//----------------------------------------------------------------------

	/**
	 * Create a new set statement.
	 * @param target the target attribute
	 * @param formula the formula to calculate
	 */
	public Stmt_Set(FormulaAttribute target, RpdlFormula formula, int op) {
		super(formula);
		// adjust the formula in the case we have "+="
		switch (op) {
			case BasicValue.OP_ADD:
				RpdlFormula f = new FormulaMath(op, target, formula);
				setFormula(f);
				break;
			case BasicValue.COMP_EQ:
				break;
			default:
				error("unsupported operation '" + BasicValue.getOpName(op) + "' for formula");
		}

		this.target = target;
	}

	//----------------------------------------------------------------------
	//------- Accessors
	//----------------------------------------------------------------------
	
	/** Get the target formula. */
	public final FormulaAttribute getTarget() {return target;}

	/** Get the priority. */
	public int getPri() {return priority;}
	/** Set a new priority. */
	public void setPri(int pri) {priority = pri;}
	
	
	//----------------------------------------------------------------------
	//------- Overridden methods of the superclass
	//----------------------------------------------------------------------
	/** Get the execution priority. This is PRI_NORM + getPri()*/
	public int getPriority() {return PRI_NORM + getPri();}
	/**
	 * This method can be overwritten, and then used to check the validity
	 * of the statement. This method is called from the addCommand() of CmdCrType & CmdCrObj
	 * The default implementaion does nothing.
	 */
	public void checkStatement(RpdlBase base, Entity e) throws RpdlException {
		super.checkStatement(base, e);
		try {
			String targetType = getTarget().getResultType(base, e);
			// add a type-cast to the formula, if necessary
			RpdlFormula f = getFormula();
			String fmlType = f.getResultType(base, e);
			if (!fmlType.equals(targetType)) {
				debug("adding cast to formula: " + f);
				f = new FormulaMath(BasicValue.getCastOperation(targetType), f, null);
				debug("cast to formula: " + f);
				setFormula(f);
			}
			// try again
			fmlType = f.getResultType(base, e);
			if (!fmlType.equals(targetType)) {
				throw new RpdlException("Targettype (" + targetType + ") doesn't match result (" + fmlType + ")" +
										" in SET-Statement '" + toString() + "'", getOrigin());
			}
		} catch (Exception ex) {
			ex.printStackTrace();
			throw new RpdlException("Error in formula " + toString() + ": "+ ex.getMessage(), getOrigin());
		}
	}
	
	/**
	 * Excecute the statement. Here the actual calculation for SET-statements
	 * are done. 
	 */
	public void execute(RpdlBase main, RpdlObject obj) throws RpdlException {
		super.execute(main, obj);
		// perform the calculation
		BasicValue res = getFormula().calc(main, obj);
		Set depSet = getTarget().getDependencies(main, obj);
		for (Iterator it=depSet.iterator(); it.hasNext(); ) {
			Object dep = it.next();
			if (dep instanceof RpdlAttribute) {
				RpdlAttribute attr = (RpdlAttribute) dep;
				attr.setValue(res);
				continue;
			}
			error("unknown target " + dep + " in formula " + getTarget() + ", " + getTarget().getOrigin());
		}
	}
	
	/**
	 * Get all the target attributes this statement will modify.
	 */
	public Set getTargets(RpdlBase main, RpdlObject obj) throws RpdlException {
		return getTarget().getDependencies(main, obj);
	}
	
	/** Get a string which specifies the priority. e.g. SET, SETFIRST, SETLAST. */
	public String getSetString() {
		switch(getPri()) {
			case SETPRI_FIRST:	return "SETFIRST";
			case SETPRI_NORM:	return "SET";
			case SETPRI_LAST:	return "SETLAST";
			default:		return "SET???";
		}
	}
	//----------------------------------------------------------------------
	//------- Overridden system methods
	//----------------------------------------------------------------------
	public String toString() {
		return "(" + getID() + ") "+getSetString()+" " + getTarget() + " = " + getFormula() + ";";
	}
}